@@ -1,15 +1,16 @@
-use std::alloc::Layout;
 use std::convert::TryFrom;
 
+use tracing::debug;
+
 use crate::{
     bytecode::{decode_str, read_from_bytes, TriviallyEncodable},
-    collections::key_map::Key,
+    collections::key_map::Handle,
     procedures::ExecutionError,
     procedures::ExecutionResult,
     program::CaoProgram,
     traits::MAX_STR_LEN,
     value::Value,
-    StrPointer, VariableId,
+    VariableId,
 };
 
 use super::{
@@ -83,7 +84,7 @@ pub fn instr_len<T>(vm: &mut Vm<T>) -> ExecutionResult {
         Value::Integer(_) | Value::Floating(_) => 1,
         Value::String(s) => {
             let st = unsafe {
-                vm.get_str(s).ok_or_else(|| {
+                s.get_str().ok_or_else(|| {
                     ExecutionError::invalid_argument("String not found".to_string())
                 })?
             };
@@ -107,20 +108,8 @@ pub fn instr_string_literal<T>(
     let payload = read_str(&mut (handle as usize), program.data.as_slice())
         .ok_or_else(|| ExecutionError::invalid_argument(None))?;
 
-    unsafe {
-        let layout = Layout::from_size_align(4 + payload.len(), 4).unwrap();
-        let mut ptr = vm
-            .runtime_data
-            .memory
-            .alloc(layout)
-            .map_err(|_| ExecutionError::OutOfMemory)?;
-
-        let result: *mut u8 = ptr.as_mut();
-        std::ptr::write(result as *mut u32, payload.len() as u32);
-        std::ptr::copy(payload.as_ptr(), result.add(4), payload.len());
-
-        vm.stack_push(Value::String(StrPointer(ptr.as_ptr())))?;
-    }
+    let ptr = vm.init_string(payload)?;
+    vm.stack_push(Value::String(ptr))?;
 
     Ok(())
 }
@@ -130,7 +119,7 @@ pub fn instr_jump(
     program: &CaoProgram,
     runtime_data: &mut RuntimeData,
 ) -> ExecutionResult {
-    let label: Key = unsafe { decode_value(&program.bytecode, instr_ptr) };
+    let label: Handle = unsafe { decode_value(&program.bytecode, instr_ptr) };
     let argcount: u32 = unsafe { decode_value(&program.bytecode, instr_ptr) };
 
     // remember the location after this jump
@@ -159,7 +148,7 @@ pub fn instr_jump(
 }
 
 pub fn execute_call<T>(vm: &mut Vm<T>, instr_ptr: &mut usize, bytecode: &[u8]) -> ExecutionResult {
-    let fun_hash: Key = unsafe { decode_value(bytecode, instr_ptr) };
+    let fun_hash: Handle = unsafe { decode_value(bytecode, instr_ptr) };
     let procedure = vm
         .callables
         .remove(fun_hash)
@@ -231,7 +220,7 @@ pub fn instr_return<T>(vm: &mut Vm<T>, instr_ptr: &mut usize) -> ExecutionResult
 }
 
 pub fn begin_repeat<T>(vm: &mut Vm<T>) -> ExecutionResult {
-    let n = vm.runtime_data.stack.pop();
+    let n = vm.runtime_data.stack.last();
     let n = i64::try_from(n).map_err(|_| {
         ExecutionError::invalid_argument("Repeat input must be an integer".to_string())
     })?;
@@ -240,24 +229,30 @@ pub fn begin_repeat<T>(vm: &mut Vm<T>) -> ExecutionResult {
             "Repeat input must be non-negative".to_string(),
         ));
     }
-    // restore the counter if it's valid, to be used by Repeat
-    vm.stack_push(n)?;
+    // init the loop counter
+    vm.stack_push(0)?;
     Ok(())
 }
 
-/// return i
-pub fn repeat<T>(vm: &mut Vm<T>) -> Result<i64, ExecutionError> {
-    let i = vm.runtime_data.stack.pop();
-    let mut i = i64::try_from(i).map_err(|_| {
-        ExecutionError::invalid_argument("Repeat input must be an integer".to_string())
-    })?;
-    i -= 1;
-    if i >= 0 {
-        // restore the variable
+/// return if the loop should continue
+pub fn repeat<T>(vm: &mut Vm<T>) -> Result<bool, ExecutionError> {
+    let i = vm.stack_pop();
+    let i = i64::try_from(i).expect("Repeat input `I` must be an integer");
+    let n = vm.runtime_data.stack.last();
+    let n = i64::try_from(n).expect("Repeat input `N` must be an integer");
+
+    let should_continue = i < n;
+    if should_continue {
+        // restore the variable and add 1
+        vm.stack_push(i + 1)?;
+        // push the lane argument
         vm.stack_push(i)?;
+    } else {
+        // clean up
+        vm.stack_pop(); // N
     }
 
-    Ok(i)
+    Ok(should_continue)
 }
 
 pub fn instr_copy_last<T>(vm: &mut Vm<T>) -> ExecutionResult {
@@ -270,43 +265,42 @@ pub fn instr_copy_last<T>(vm: &mut Vm<T>) -> ExecutionResult {
     Ok(())
 }
 
-/// push N and I onto the stack
-/// I shall be counted down (N-1..0)
+/// push `i=0` onto the stack
 pub fn begin_for_each<T>(vm: &mut Vm<T>) -> ExecutionResult {
     let item = vm.runtime_data.stack.last();
-    if !item.is_obj() {
-        return Err(ExecutionError::invalid_argument(
-            "ForEach is only valid for tables".to_string(),
-        ));
-    }
+    // test if the input is a table
+    let _item = vm.get_table(item)?;
+    debug!("Starting for-each on table {:?}", _item);
 
-    let item = vm.get_table(item)?;
-    let n = item.len() as i64;
-
-    vm.stack_push(n)?;
-    vm.stack_push(0)?;
+    vm.stack_push(0)?; // i, this should be incremented by `for_each`
 
     Ok(())
 }
 
+/// Assumes that [begin_for_each](begin_for_each) was called once to set up the loop
+///
+/// Requires `i`, and the object to be on the stack.
+///
+/// Pushes the next key and the object onto the stack. Assumes that the lane takes these as
+/// parameters.
 pub fn for_each<T>(vm: &mut Vm<T>) -> Result<bool, ExecutionError> {
     let i = vm.stack_pop();
-    let n = vm.runtime_data.stack.peek_last(0);
-    let obj_val = vm.runtime_data.stack.peek_last(1);
+    let obj_val = vm.runtime_data.stack.peek_last(0);
 
     let mut i = i64::try_from(i).expect("Repeat input #0 must be an integer");
-    let n = i64::try_from(n).expect("Repeat input #1 must be an integer");
     let obj = vm
         .get_table(obj_val)
         .expect("Repeat input #2 must be a table");
 
-    debug_assert!(0 <= i);
+    debug_assert!(0 <= i, "for_each overflow");
 
-    let should_continue = i < n;
+    let n = obj.len() as i64;
+
+    let should_continue = 0 <= i && i < n;
     if should_continue {
         let key = obj.iter().nth(i as usize).map(|(k, _)| k).ok_or_else(|| {
             ExecutionError::invalid_argument(format!(
-                "ForEach can not be completed. i: {} n: {}",
+                "ForEach can not find the `i`th argument. i: {} n: {}\nDid you remove items during iteration?",
                 i, n
             ))
         })?;
@@ -318,9 +312,7 @@ pub fn for_each<T>(vm: &mut Vm<T>) -> Result<bool, ExecutionError> {
         vm.stack_push(key)?;
         vm.stack_push(obj_val)?;
     } else {
-        vm.stack_pop(); // clean up N
         vm.stack_pop(); // clean up the object
-        vm.stack_push(false)?; // for the next GotoIfTrue
     }
 
     Ok(should_continue)
